/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - solarisMib2.c */
/* rdk@krupczak.org */

/* a plugin implementing MIB-2 objects for Solaris */
/* we support IPv4 and UDP6 and TCP6 */

#ifdef SOLARIS

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stropts.h>
#include <tiuser.h>
#include <sys/conf.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <strings.h>
#include <dlfcn.h>
#include <dirent.h>
#include <sys/tihdr.h>
#include <sys/tpicommon.h>
#include <errno.h>

#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <inet/mib2.h>
#include <inet/ip.h>
#include <net/if_arp.h>
#include <net/route.h>
#include <sys/sockio.h>
#include <assert.h>

#include "plugins.h"
#include "xmp.h"
#include "mib2plugin.h"
#include "socklib.h"

/* ************************************************************* */
/* plugin globals, data structures, defines, etc. */

static pthread_t wt;
static int s; /* for Solaris MIB2-interface */
extern int mib2ThreadKill;

int get_named_kstat(char *, int, char *, char *, void *, int);

/* our platform-independent structures */

int debugLevel = 0;
time_t statsTime = 0;
ipStats_t ipstats;
tcpStats_t tcpstats;
icmpStats_t icmpstats;
udpStats_t udpstats;

mib2UdpEndpt_t *udpTable = NULL;
int numUdps = 0, maxUdps = 0;
time_t udpTableTime = 0;

arpEntry_t *arpTable = NULL;
int numArps = 0, maxArps = 0;
time_t arpTableTime = 0;

mib2IfEntry_t *ifTable = NULL;
int numIfs = 0, maxIfs = 0;
struct lifreq *lifreqs = NULL;
time_t ifTableTime = 0;

ipRoute_t *routeTable = NULL;
int numRoutes = 0, maxRoutes = 0;
time_t routeTableTime = 0;

ipAddrEntry_t *addrTable = NULL;
int numAddrs = 0, maxAddrs = 0;
time_t ipAddrTableTime = 0;

mib2TcpConn_t *tcpTable = NULL;
int numTcpConnections = 0, maxTcpConnections = 0;
time_t tcpTableTime = 0;

static int solaris_udpEntrySize = 0; /* entry sizes passed in stats */
static int solaris_tcpEntrySize = 0; /* structs for particular proto */
static int solaris_ipAddrEntrySize = 0;
static int solaris_ipRouteEntrySize = 0;
static int solaris_ipNetToMediaEntrySize = 0;
static int solaris_udp6EntrySize = 0;
static int solaris_tcp6EntrySize = 0; 

pthread_mutex_t mib2_lock = PTHREAD_MUTEX_INITIALIZER;

/* ************************************************************* */




/* ************************************************************* */
/* caller should already have lock for table */

static int enlarge_udp_table()
{
     /* if needing initialization */
     if (udpTable == NULL) {
        udpTable = (mib2UdpEndpt_t *)malloc(sizeof(mib2UdpEndpt_t) * 50);
        if (udpTable == NULL) {
	   perror("enlarge_udp_table: initial malloc failed");
           return -1;
        }
        maxUdps = 50;
     }

     if (maxUdps == numUdps) {
        maxUdps += 25;
        udpTable = (mib2UdpEndpt_t *)realloc(udpTable,
		    maxUdps*sizeof(mib2UdpEndpt_t));

        if (udpTable == NULL) {
	   perror("enlarge_udp_table: realloc failed");
           maxUdps = numUdps = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_route_table()
{
     /* if needing initialization */
     if (routeTable == NULL) {
        routeTable = (ipRoute_t *)malloc(sizeof(ipRoute_t) * 5);
        if (routeTable == NULL) {
	   perror("enlarge_route_table: initial malloc failed");
           return -1;
        }
        maxRoutes = 5;
     }

     if (maxRoutes == numRoutes) {
        maxRoutes += 5;
        routeTable = (ipRoute_t *)realloc(routeTable,
		    maxRoutes*sizeof(ipRoute_t));

        if (routeTable == NULL) {
	   perror("enlarge_route_table: realloc failed");
           maxRoutes = numRoutes = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_addr_table()
{
     /* if needing initialization */
     if (addrTable == NULL) {
        addrTable = (ipAddrEntry_t *)malloc(sizeof(ipAddrEntry_t) * 5);
        if (addrTable == NULL) {
	   perror("enlarge_addr_table: initial malloc failed");
           return -1;
        }
        maxAddrs = 5;
     }

     if (maxAddrs == numAddrs) {
        maxAddrs += 5;
        addrTable = (ipAddrEntry_t *)realloc(addrTable,
		    maxAddrs*sizeof(ipAddrEntry_t));

        if (addrTable == NULL) {
	   perror("enlarge_addr_table: realloc failed");
           maxAddrs = numAddrs = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */

static int enlarge_if_table()
{
     /* if needing initialization */
     if (ifTable == NULL) {
        ifTable = (mib2IfEntry_t *)malloc(sizeof(mib2IfEntry_t) * 10);
        if (ifTable == NULL) {
	   perror("enlarge_if_table: initial malloc failed");
           return -1;
        }
        maxIfs = 10;

        lifreqs = (struct lifreq *)malloc(sizeof(struct lifreq)*maxIfs);
        if (lifreqs == NULL) {
	   perror("enlarge_if_table: initial malloc of lifreqs failed");
           maxIfs = 0;
        }
        bzero((void *)lifreqs,sizeof(struct lifreq)*maxIfs);
     }

     if (maxIfs == numIfs) {
        maxIfs += 5;
        ifTable = (mib2IfEntry_t *)realloc(ifTable,
		    maxIfs*sizeof(mib2IfEntry_t));

        if (ifTable == NULL) {
	   perror("enlarge_if_table: realloc failed");
           maxIfs = numIfs = 0;
           return -1;
        }

        lifreqs = (struct lifreq *)realloc(lifreqs,
                                           maxIfs*sizeof(struct lifreq));
        if (lifreqs == NULL) {
           maxIfs = numIfs = 0;
           return -1;
        }
     }

     return 1;
}

/* ************************************************************* */

static int enlarge_arp_table()
{
     /* if needing initialization */
     if (arpTable == NULL) {
        arpTable = (arpEntry_t *)malloc(sizeof(arpEntry_t) * 5);
        if (arpTable == NULL) {
	   perror("enlarge_arp_table: initial malloc failed");
           return -1;
        }
        maxArps = 5;
     }

     if (maxArps == numArps) {
        maxArps += 5;
        arpTable = (arpEntry_t *)realloc(arpTable,
		    maxArps*sizeof(arpEntry_t));

        if (arpTable == NULL) {
	   perror("enlarge_arp_table: realloc failed");
           maxArps = numArps = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */
/* caller should already have lock for table */

static int enlarge_tcp_table()
{
     /* if needing initialization */
     if (tcpTable == NULL) {
        tcpTable = (mib2TcpConn_t *)malloc(sizeof(mib2TcpConn_t) * 50);
        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: initial malloc failed");
           return -1;
        }
        maxTcpConnections = 50;
     }

     if (maxTcpConnections == numTcpConnections) {
        maxTcpConnections += 25;
        tcpTable = (mib2TcpConn_t *)realloc(tcpTable,
		    maxTcpConnections*sizeof(mib2TcpConn_t));

        if (tcpTable == NULL) {
	   perror("enlarge_tcp_table: realloc failed");
           maxTcpConnections = numTcpConnections = 0;
           return -1;
        }

     }

     return 1;
}

/* ************************************************************* */
/* ************************************************************* */

static void reloadIfTable()
{
     static int sock = 0;
     static int sock6 = 0;
     int i,fd,j,ret;
     struct lifnum l;
     struct lifconf lifconf;
     struct lifreq lif;
     struct sockaddr_in *sa;
     struct sockaddr_in6 *sa6;
     char devname[64];
     uint64_t li;
     struct arpreq arp;
     int numLoopbacks, numNetIfs;
     static int firstTime = 1;

     if (sock == 0) {
       if ((sock = socket(AF_INET,SOCK_DGRAM,IPPROTO_IP)) < 0) { 
           perror("reloadIfTable: socket call failed");
           return;
        }
       if ((sock6 = socket(AF_INET6,SOCK_DGRAM,IPPROTO_IP)) < 0) { 
           perror("reloadIfTable: socket6 call failed");
           return;
        }
     }

     /* get lock on ifTable */
     pthread_mutex_lock(&mib2_lock);

     /* we ask for all interfaces (IPv4 and IPv6) and they come
        out without really any signifier (as far as I can tell)
        as to which they are; sooooo
  
        first time we see an interface, we spec it as IPv4,
        and the second time we set it as IPv6
     */

     /* get number interfaces */
     l.lifn_family = AF_INET;  /* only IPv4 interfaces */
     l.lifn_family = AF_UNSPEC;  /* both interfaces */
     l.lifn_flags = 0;
     if (ioctl(sock,SIOCGLIFNUM,&l) < 0) {
        perror("reloadIfTable: SIOCGLIFNUM v4 failed");
        numIfs = 0;
     }
     numIfs = l.lifn_count;

     if (debugLevel)
        printf("reloadIfTable: %d interfaces\n",numIfs);

     enlarge_if_table();

     /* do SIOCGLIFCONF to get interface structs */
     /* XXX look for example code to get if_msghdr_t and if_data */
     /* in net/if.h */
     /* see also if_tcp(7) man page */
     /* streamio man page */
     /* dlpi man page */

     /* do SIOCGLIFCONF */
     bzero((void *)&lifconf,sizeof(lifconf));
     lifconf.lifc_family = AF_INET;  /* only IPv4 interfaces */
     lifconf.lifc_family = AF_UNSPEC;  /* get both ipv4 and ipv6 */
     lifconf.lifc_flags = 0;
     lifconf.lifc_buf = (caddr_t)lifreqs;
     lifconf.lifc_len = sizeof(struct lifreq)*maxIfs;
     bzero((void *)lifreqs,sizeof(struct lifreq)*maxIfs);
     if ((ret = ioctl(sock,SIOCGLIFCONF,(char *)&lifconf)) < 0) {
        perror("SIOCGLIFCONF failed");
        numIfs = 0;
        printf("lifreqs is %lx\n",lifreqs);
        pthread_mutex_unlock(&mib2_lock);
        return;
     }

     /* now, for each interface we find, do the ioctls */
     numLoopbacks = numNetIfs = 0;
     
     for (i=0; i<numIfs; i++) {

         enlarge_if_table();

         /* first time through, bzero out structure; afterwards, we're
            making the assumption that the interfaces are going to
            come out of the kernel in the same order each time;
            long term, we should cache the old oper status in a 
            separate array/structure XXX */
         if (firstTime)
            bzero((void *)&ifTable[i],sizeof(mib2IfEntry_t));

         strcpy(ifTable[i].ifname,lifreqs[i].lifr_name);

         DBG1("reloadIfTable: found '%s'\n",ifTable[i].ifname);

         /* SIOCGLIFFLAGS */
         sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
         if (ioctl(sock,SIOCGLIFFLAGS,&lif) < 0)
 	    perror("siocglifflags failed");
         else {
	    ifTable[i].lflags = lif.lifr_flags;
         }

         /* given flags, determine ifAdminStatus and ifOperStatus */
   
         /* dont forget to set ifIndex even though we dont use it
            for XMP other than link Up/Down Traps */

         ifTable[i].ifIndex = i+1; 

         ifTable[i].oldIfOperStatus = ifTable[i].ifOperStatus;

         ifTable[i].ifOperStatus = 2;
         if (ifTable[i].lflags & IFF_RUNNING)
            ifTable[i].ifOperStatus = 1;
         ifTable[i].ifAdminStatus = 2;
         if (ifTable[i].flags & IFF_UP)
            ifTable[i].ifOperStatus = 1; 

         /* assume its an IPv4 address */
         /* SIOCGLIFADDR (AF_INET=2 addr) */
         sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
         if (ioctl(sock,SIOCGLIFADDR,&lif) < 0) {
	    perror("siocglifaddr failed");
         }
         else {
            sa = (struct sockaddr_in *)&lif.lifr_addr;
            if (sa->sin_family == AF_INET) {
               DBG1("Got an IPv4/sock address for '%s'\n",ifTable[i].ifname);
	       ifTable[i].ipVersion = 4;
	       ifTable[i].ipv4addr = sa->sin_addr.s_addr;
            }
         }

         /* if this interface is already in the list, then
            try to see if we have an IPv6 address for it */
         for (j=0; j<i; j++) {

  	     /* if we find that this interface is already in the 
                table, make this interface be a IPv6 addr */
	     if (strcmp(ifTable[i].ifname,ifTable[j].ifname) == 0) {

	        DBG1("Looking for IPv6 address for '%s'\n",
                      ifTable[i].ifname);

                sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
                if (ioctl(sock6,SIOCGLIFADDR,&lif) < 0) {
  	           perror("siocglifaddr failed");
                }

                if (sa->sin_family == AF_INET6) {
  	           sa6 = (struct sockaddr_in6 *)&lif.lifr_addr;
	           ifTable[i].ipVersion = 6;
	           ifTable[i].ipv4addr = 0; 
                   convertIPv6Addr(&sa6->sin6_addr,
                        ifTable[i].ipv6Addr,
	                sizeof(ifTable[i].ipv6Addr));
                   DBG2("Got an IPv6 addr '%s' for '%s'\n",
                        ifTable[i].ipv6Addr,
                        ifTable[i].ifname);

                }
 
   	     } /* if interface is already in table */

         } /* for j=0; j<i */

         /* SIOCGLIFNETMASK get netmask */
         sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
         if (ioctl(sock,SIOCGLIFNETMASK,&lif) < 0) {
	    perror("siocglifnetmask failed");
         }
         else {
            sa = (struct sockaddr_in *)&lif.lifr_addr;
            if (sa->sin_family == AF_INET) {
	       ifTable[i].ipv4mask = sa->sin_addr.s_addr;
            }
            if (sa->sin_family == AF_INET6) {
	       ifTable[i].ipv4mask = 0;
            }
         }

         /* SIOCGLIFMTU */
         sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
         if (ioctl(sock,SIOCGLIFMTU,&lif) < 0) {
	    perror("siocglifmtu failed");
         }
         else {
            ifTable[i].mtu = lif.lifr_mtu;
         }

         /* SIOCGLIFMETRIC ? */

         /* SIOCGLIFBRDADDR ? */
#if 0
         this IOCTL seems to fail despite what include files
	 say.  For now, dont bother.  
         sprintf(lif.lifr_name,"%s",ifTable[i].ifname);
         if (ioctl(sock,SIOCGLIFBRDADDR,&lif) < 0) {
	    perror("siocglifbrdaddr failed");
         }
#endif

         /* SIOCGLIFLNKINFO ? */

         /* for other data, use either kstat or if_data struct */
         ifTable[i].ifType = 1;  /* other(1) */
         ifTable[i].ifSpeed = 1; /* for now XXX */
         ifTable[i].alias[0] = 0; /* init default */

         /* XXX yuck!  there has to be a more automated way to determine
            ifType */

         /* figure out aliases for easier polling */

         if (strncmp(ifTable[i].ifname,"lo",2) == 0) {
      	    ifTable[i].ifType = 24; /* softwareloopback(24) */
            sprintf(ifTable[i].alias,"lo%d",numLoopbacks);
            numLoopbacks++;
         }
         else {
            sprintf(ifTable[i].alias,"if%d",numNetIfs);
            numNetIfs++;
         }

         if (strncmp(ifTable[i].ifname,"tr",2) == 0) {
      	    ifTable[i].ifType = 9;  /* tokenring(9) */
         }
         if (strncmp(ifTable[i].ifname,"ppp",3) == 0) {
 	    ifTable[i].ifType = 23;  /* ppp(23) */
         }
         if (strncmp(ifTable[i].ifname,"sl",2) == 0) {
 	    ifTable[i].ifType = 28;  /* slip(28) */
         }
         if (strncmp(ifTable[i].ifname,"eri",3) == 0) {
 	    ifTable[i].ifType = 62; /* fastether(62) */
         }
         if (strncmp(ifTable[i].ifname,"le",2) == 0) {
	    ifTable[i].ifType = 6; /* ether(6) */
         }
         if (strncmp(ifTable[i].ifname,"elxl",4) == 0) {
	    ifTable[i].ifType = 62; /* fastether(62) */
         }
         if (strncmp(ifTable[i].ifname,"hme",3) == 0) {
	    ifTable[i].ifType = 62; /* fastether(62) */
         }

         if (debugLevel)
            printf("Interface: %s , mtu:%d\n",ifTable[i].ifname,
                   ifTable[i].mtu);

         DBG2("reloadIfTable: %s aliased to %s\n",
	      ifTable[i].ifname,
	      ifTable[i].alias);

#if 0
         /* to get rest of info for interface, we can
            open up /dev/interfacename (w/o trailing physical unit) */
         sprintf(devname,"/dev/%s",ifTable[i].ifname);
         if ((fd = open(devname,O_RDWR,0600)) > 0) {
	    close(fd);
         }
         /* strip off physical unit */
         sprintf(devname,"/dev/%s",ifTable[i].ifname);
         for (j=0; j<strlen(devname); j++) {
	     if (isdigit(devname[j])) {
	        devname[j] = 0;
                break;
             }
         }
         if ((fd = open(devname,O_RDWR,0600)) > 0) {
	   close(fd); /* XXX */
         }
#endif
         /* we cant get our physical interface addr via arp table */
         sa = (struct sockaddr_in *)&arp.arp_pa;
         sa->sin_family = AF_INET;
         sa->sin_port = 0;
         sa->sin_addr.s_addr = ifTable[i].ipv4addr;
         arp.arp_pa.sa_family = AF_INET;
         if (ioctl(sock,SIOCGARP,&arp) >= 0) {
	    bcopy(arp.arp_ha.sa_data,ifTable[i].hwaddr,6);
            ifTable[i].hwaddrlen = 6;
         }

         /* need to use kstat to get interface counters */
         get_named_kstat(NULL,0,ifTable[i].ifname,"ifspeed",
                         &ifTable[i].ifSpeed,sizeof(ifTable[i].ifSpeed));
         get_named_kstat(NULL,0,ifTable[i].ifname,"ipackets",
                         &ifTable[i].ifInUcastPkts,
                         sizeof(ifTable[i].ifInUcastPkts));
         get_named_kstat(NULL,0,ifTable[i].ifname,"rbytes",
                         &ifTable[i].ifInOctets,
                         sizeof(ifTable[i].ifInOctets));
         get_named_kstat(NULL,0,ifTable[i].ifname,"ierrors",
                         &ifTable[i].ifInErrors,
                         sizeof(ifTable[i].ifInErrors));

         get_named_kstat(NULL,0,ifTable[i].ifname,"opackets",
                         &ifTable[i].ifOutUcastPkts,
                         sizeof(ifTable[i].ifOutUcastPkts));
         get_named_kstat(NULL,0,ifTable[i].ifname,"obytes",
                         &ifTable[i].ifOutOctets,
                         sizeof(ifTable[i].ifOutOctets));
         get_named_kstat(NULL,0,ifTable[i].ifname,"oerrors",
                         &ifTable[i].ifOutErrors,
                         sizeof(ifTable[i].ifOutErrors));

         get_named_kstat(NULL,0,ifTable[i].ifname,"brdcstxmt",
                         &ifTable[i].ifOutNUcastPkts,
                         sizeof(ifTable[i].ifOutNUcastPkts));
         li = 0;
         get_named_kstat(NULL,0,ifTable[i].ifname,"multxmt",
                         &li,
                         sizeof(li));
         ifTable[i].ifOutNUcastPkts += li;

         get_named_kstat(NULL,0,ifTable[i].ifname,"brdcstrcv",
                         &ifTable[i].ifInNUcastPkts,
                         sizeof(ifTable[i].ifInNUcastPkts));
         li = 0;
         get_named_kstat(NULL,0,ifTable[i].ifname,"multrcv",
                         &li,
                         sizeof(li));
         ifTable[i].ifInNUcastPkts += li;

         get_named_kstat(NULL,0,ifTable[i].ifname,"norcvbuf",&li,sizeof(li));
         ifTable[i].ifInDiscards += li;
         get_named_kstat(NULL,0,ifTable[i].ifname,"drop",&li,sizeof(li));
         ifTable[i].ifInDiscards += li;

         get_named_kstat(NULL,0,ifTable[i].ifname,"noxmtbuf",&li,sizeof(li));
         ifTable[i].ifOutDiscards += li;
         get_named_kstat(NULL,0,ifTable[i].ifname,"nocanput",&li,sizeof(li));
         ifTable[i].ifOutDiscards += li;

         /* if this is not our first time, compare old and new ifOperStatus */
         if (firstTime != 1) {

	    if (ifTable[i].oldIfOperStatus != ifTable[i].ifOperStatus) {

	       if (ifTable[i].ifOperStatus == 2) {
	          sendLinkDownTrap(ifTable[i].ifIndex,
			           ifTable[i].ifAdminStatus,
				   ifTable[i].ifOperStatus,
                                   ifTable[i].ifname);
	       }
               if (ifTable[i].ifOperStatus == 1) {
	          sendLinkUpTrap(ifTable[i].ifIndex,
			         ifTable[i].ifAdminStatus,
			         ifTable[i].ifOperStatus,
                                 ifTable[i].ifname);
               }

	    } /* if old != new */

         } /* if !firstTime */

     } /* for each IPv4 if */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     ifTableTime = time(NULL);
     firstTime = 0;

     return;
}

/* ************************************************************* */
/* ************************************************************* */

static void convert_icmpstats(mib2_icmp_t *s, icmpStats_t *d)
{
     if ((s == NULL) || (d == NULL))   
        return;

     d->icmpInMsgs = (u_int64_t)s->icmpInMsgs;
     d->icmpInErrors = (u_int64_t)s->icmpInErrors;
     d->icmpInDestUnreachs = (u_int64_t)s->icmpInDestUnreachs;
     d->icmpInTimeExcds = (u_int64_t)s->icmpInTimeExcds;
     d->icmpInParmProbs = (u_int64_t)s->icmpInParmProbs;
     d->icmpInSrcQuenchs = (u_int64_t)s->icmpInSrcQuenchs;
     d->icmpInRedirects = (u_int64_t)s->icmpInRedirects;
     d->icmpInEchos = (u_int64_t)s->icmpInEchos;
     d->icmpInEchoReps = (u_int64_t)s->icmpInEchoReps;
     d->icmpInTimestamps = (u_int64_t)s->icmpInTimestamps;
     d->icmpInTimestampReps = (u_int64_t)s->icmpInTimestampReps;
     d->icmpInAddrMasks = (u_int64_t)s->icmpInAddrMasks;
     d->icmpInAddrMaskReps = (u_int64_t)s->icmpInAddrMaskReps;
     d->icmpOutMsgs = (u_int64_t)s->icmpOutMsgs;
     d->icmpOutErrors = (u_int64_t)s->icmpOutErrors;
     d->icmpOutDestUnreachs = (u_int64_t)s->icmpOutDestUnreachs;
     d->icmpOutTimeExcds = (u_int64_t)s->icmpOutTimeExcds;
     d->icmpOutParmProbs = (u_int64_t)s->icmpOutParmProbs;
     d->icmpOutSrcQuenchs = (u_int64_t)s->icmpOutSrcQuenchs;
     d->icmpOutRedirects = (u_int64_t)s->icmpOutRedirects;
     d->icmpOutEchos = (u_int64_t)s->icmpOutEchos;
     d->icmpOutEchoReps = (u_int64_t)s->icmpOutEchoReps;
     d->icmpOutTimestamps = (u_int64_t)s->icmpOutTimestamps;
     d->icmpOutTimestampReps = (u_int64_t)s->icmpOutTimestampReps;
     d->icmpOutAddrMasks = (u_int64_t)s->icmpOutAddrMasks;
     d->icmpOutAddrMaskReps = (u_int64_t)s->icmpOutAddrMaskReps;

     if (debugLevel) {
        printf("convert_icmpstats: icmp in:%lld out:%lld\n",
               d->icmpInMsgs, d->icmpOutMsgs);
        printf("convert_icmpstats: icmp in:%ld out:%ld\n",
               s->icmpInMsgs, s->icmpOutMsgs);
     }

     return;
}

/* ************************************************************* */

static void convert_ipstats(mib2_ip_t *s, ipStats_t *d)  
{ 
     if ((s == NULL) || (d == NULL))   
        return;

     d->ipForwarding = (int)       s->ipForwarding;
     d->ipDefaultTTL = (int)       s->ipDefaultTTL;
     d->ipInReceives = (u_int64_t) s->ipInReceives;
     d->ipInHdrErrors = (u_int64_t) s->ipInHdrErrors;
     d->ipInAddrErrors = (u_int64_t) s->ipInAddrErrors;
     d->ipForwDatagrams = (u_int64_t) s->ipForwDatagrams;
     d->ipInUnknownProtos = (u_int64_t) s->ipInUnknownProtos;
     d->ipInDiscards = (u_int64_t) s->ipInDiscards;
     d->ipInDelivers = (u_int64_t) s->ipInDelivers;
     d->ipOutRequests = (u_int64_t) s->ipOutRequests;
     d->ipOutDiscards = (u_int64_t) s->ipOutDiscards;
     d->ipOutNoRoutes = (u_int64_t) s->ipOutNoRoutes;
     d->ipReasmTimeout = (u_int64_t) s->ipReasmTimeout;
     d->ipReasmReqds = (u_int64_t) s->ipReasmReqds;
     d->ipReasmOKs = (u_int64_t) s->ipReasmOKs;
     d->ipReasmFails = (u_int64_t) s->ipReasmFails;
     d->ipFragOKs = (u_int64_t) s->ipFragOKs;
     d->ipFragFails = (u_int64_t) s->ipFragFails;
     d->ipFragCreates = (u_int64_t) s->ipFragCreates;
     d->ipRoutingDiscards = (u_int64_t) s->ipRoutingDiscards;

     /* dont forget the udp/tcp stats stashed in IP stats */
     udpstats.udpNoPorts = s->udpNoPorts;
     tcpstats.tcpInErrs = s->tcpInErrs;

     solaris_ipAddrEntrySize = s->ipAddrEntrySize;
     solaris_ipRouteEntrySize = s->ipRouteEntrySize;
     solaris_ipNetToMediaEntrySize = s->ipNetToMediaEntrySize;

     if (debugLevel)
        printf("convert_ipstats: forw:%d inRecv:%lld OutReq:%lld\n",
               d->ipForwarding,d->ipInReceives,d->ipOutRequests);

     return; 
}

/* ************************************************************* */
/* ************************************************************* */

static void convert_udpstats(mib2_udp_t *s, udpStats_t *d) 
{ 
     if ((s == NULL) || (d == NULL))   
        return;

     d->udpInDatagrams = (u_int64_t) s->udpInDatagrams;

     /* udpNoPorts is within mib2_ip_t */

     d->udpInErrors = (u_int64_t) s->udpInErrors;
     d->udpOutDatagrams = (u_int64_t) s->udpOutDatagrams;

     solaris_udpEntrySize = s->udpEntrySize;
     solaris_udp6EntrySize = s->udp6EntrySize;

     if (debugLevel)
        printf("convert_udpstats: in:%lld out:%lld\n",
               d->udpInDatagrams, d->udpOutDatagrams);

     return; 
}

static void convert_tcpstats(mib2_tcp_t *s, tcpStats_t *d) 
{ 
     if ((s == NULL) || (d == NULL))   
        return;

     d->tcpRtoAlgorithm = (u_int64_t)s->tcpRtoAlgorithm;
     d->tcpRtoMin = (u_int64_t)s->tcpRtoMin;
     d->tcpRtoMax = (u_int64_t)s->tcpRtoMax;

     d->tcpMaxConn = (int)s->tcpMaxConn;

     d->tcpActiveOpens = (u_int64_t)s->tcpActiveOpens;
     d->tcpPassiveOpens = (u_int64_t)s->tcpPassiveOpens;
     d->tcpAttemptFails = (u_int64_t)s->tcpAttemptFails;
     d->tcpEstabResets = (u_int64_t)s->tcpEstabResets;
     d->tcpCurrEstab = (u_int64_t)s->tcpCurrEstab;
     d->tcpInSegs = (u_int64_t)s->tcpInSegs;
     d->tcpOutSegs = (u_int64_t)s->tcpOutSegs;
     d->tcpRetransSegs = (u_int64_t)s->tcpRetransSegs;
     /* d->tcpInErrs defined in ip stats !! */
     d->tcpOutRsts = (u_int64_t)s->tcpOutRsts;

     solaris_tcpEntrySize = s->tcpConnTableSize; /* misnamed */
     solaris_tcp6EntrySize = s->tcp6ConnTableSize;

     if (debugLevel)
        printf("convert_tcpstats: activeOpens:%lld curEstab:%lld\n",
               d->tcpActiveOpens, d->tcpCurrEstab);

     return;
}

/* ************************************************************* */
/* ************************************************************* */
/* assume mib2_lock already held */

static int convert_addrtable(mib2_ipAddrEntry_t *a, int buflen)
{
     int i;
     int numEntries;

     numEntries = buflen / sizeof(mib2_ipAddrEntry_t);

     DBG1("convert_addrtable: %d entries\n",numEntries);
  
     enlarge_addr_table();

     i = 0;
     while (numAddrs < numEntries) {    

         /* zero terminate the ifname */
         a[i].ipAdEntIfIndex.o_bytes[a[i].ipAdEntIfIndex.o_length] = 0;

         sprintf(addrTable[numAddrs].ifname,"%s",
                 a[i].ipAdEntIfIndex.o_bytes);

         addrTable[numAddrs].ipVersion = 4;
         addrTable[numAddrs].ipv4addr = a[i].ipAdEntAddr;
         addrTable[numAddrs].ipv4mask = a[i].ipAdEntNetMask;
         addrTable[numAddrs].ipv4bcast = a[i].ipAdEntBcastAddr;
         addrTable[numAddrs].ipv4ReasmMaxSize = 
                                       a[i].ipAdEntReasmMaxSize;

         numAddrs++;
         i++;

         enlarge_addr_table();
     }

     ipAddrTableTime = time(NULL);
}

static int convert_addr6table(mib2_ipv6AddrEntry_t *a6, int buflen)
{
     int i;
     int numEntries;

     numEntries = buflen / sizeof(mib2_ipv6AddrEntry_t);

     DBG1("convert_addr6table: %d entries\n",numEntries);
  
     enlarge_addr_table();

     i = 0;
     while (i < numEntries) {    

         /* zero terminate the ifname */
         a6[i].ipv6AddrIfIndex.o_bytes[a6[i].ipv6AddrIfIndex.o_length] = 0;

         addrTable[numAddrs].ipVersion = 6;

         sprintf(addrTable[numAddrs].ifname,"%s",
                 a6[i].ipv6AddrIfIndex.o_bytes);

         addrTable[numAddrs].ipv4addr = 0;
         addrTable[numAddrs].ipv4mask = 0;
         addrTable[numAddrs].ipv4bcast = 0;
         addrTable[numAddrs].ipv4ReasmMaxSize = 0;

         convertIPv6Addr(&a6[i].ipv6AddrAddress,
                         addrTable[numAddrs].ipv6Addr,
                         48);
         addrTable[numAddrs].ipv6PrefixLen = a6[i].ipv6AddrPfxLength;
         addrTable[numAddrs].ipv6AddrType = a6[i].ipv6AddrType;

         numAddrs++;
         i++;

         enlarge_addr_table();
     }

     ipAddrTableTime = time(NULL);
}

/* ************************************************************* */
/* ************************************************************* */

static void convert_mediatable(mib2_ipNetToMediaEntry_t *n, int buflen)
{
     int numEntries;

     numEntries = buflen / sizeof(mib2_ipNetToMediaEntry_t);

     DBG2("convert_mediatable: %d entries, buflen %d\n",numEntries,buflen);
  
     enlarge_arp_table();
     numArps = 0;

     while (numArps < numEntries) {    

         /* zero terminate the ifname */
         n[numArps].ipNetToMediaIfIndex.o_bytes[n[numArps].ipNetToMediaIfIndex.o_length] = 0;

         sprintf(arpTable[numArps].ifname,"%s",
                 n[numArps].ipNetToMediaIfIndex.o_bytes);
         arpTable[numArps].ipv4addr = n[numArps].ipNetToMediaNetAddress;

	 /* XXX still occasionally core dumping here */
         /* 10/15/2004 -- need to set hwaddrlen *before* we reference it */

         arpTable[numArps].hwaddrlen = n[numArps].ipNetToMediaPhysAddress.o_length;
         arpTable[numArps].hwtype = 0; /* not supported in solaris */
         arpTable[numArps].flags = n[numArps].ipNetToMediaInfo.ntm_flags;

         DBG2("covert_mediatable: n is %lx, numArps is %d\n",n,numArps);
         DBG3("covert_mediatable: copying from %lx to %lx, %d bytes\n",
                n[numArps].ipNetToMediaPhysAddress.o_bytes,
                arpTable[numArps].hwaddr,
                arpTable[numArps].hwaddrlen);

         bcopy(n[numArps].ipNetToMediaPhysAddress.o_bytes,
               arpTable[numArps].hwaddr,
               arpTable[numArps].hwaddrlen);
         DBG("convert_mediaentry: bcopy worked\n");

         if (debugLevel) {
            printf("arpEntry: %lx hwlen:%d %s\n",
                   arpTable[numArps].ipv4addr,
                   arpTable[numArps].hwaddrlen,
                   arpTable[numArps].ifname);
         }

         numArps++;

         enlarge_arp_table();
     }

     arpTableTime = time(NULL);
}

/* ************************************************************* */
/* ************************************************************* */

static void convert_tcp6table(mib2_tcp6ConnEntry_t *t6, int buflen)
{
     int numEntries,i;

     numEntries = buflen / sizeof(mib2_tcp6ConnEntry_t);

     DBG1("convert_tcp6table: %d entries\n",numEntries);
  
     enlarge_tcp_table();

     i = 0;
     while (i < numEntries) {

       if (t6[i].tcp6ConnLocalPort != 0) {

 	  /* ports in host byte order */

          tcpTable[numTcpConnections].ipVersion = 6;

          convertIPv6Addr(&t6[i].tcp6ConnLocalAddress,
                          tcpTable[numTcpConnections].localIPv6,48);
          tcpTable[numTcpConnections].localPort = 
                                (u_short)t6[i].tcp6ConnLocalPort;

          convertIPv6Addr(&t6[i].tcp6ConnRemAddress,
                          tcpTable[numTcpConnections].remIPv6,48);
          tcpTable[numTcpConnections].remPort = 
                                (u_short)t6[i].tcp6ConnRemPort;

          /* already in mib2 std values */
          tcpTable[numTcpConnections].state = 
                                t6[i].tcp6ConnState;

          if (debugLevel) { 
             printf("tcp6Conn: %s:%d <-> %s:%d\n",
      	            tcpTable[numTcpConnections].localIPv6,
 	            tcpTable[numTcpConnections].localPort,
	            tcpTable[numTcpConnections].remIPv6,
	            tcpTable[numTcpConnections].remPort);
          }
          numTcpConnections++;
          enlarge_tcp_table();

       } /* if localPort != 0 */
       i++;
     }

     tcpTableTime = time(NULL);

     DBG1("convert_tcp6table: %d total entries now\n",numTcpConnections);
}

static void convert_tcptable(mib2_tcpConnEntry_t *t, int buflen)
{
     int numEntries,i;

     numEntries = buflen / sizeof(mib2_tcpConnEntry_t);

     DBG1("convert_tcptable: %d entries\n",numEntries);
  
     enlarge_tcp_table();

     i = 0;
     while (i < numEntries) {

       if (t[i].tcpConnLocalPort != 0) {

 	  /* ports in host byte order */

          tcpTable[numTcpConnections].ipVersion = 4;

          tcpTable[numTcpConnections].localIp = 
                                t[i].tcpConnLocalAddress;

          tcpTable[numTcpConnections].localPort = 
                                (u_short)t[i].tcpConnLocalPort;

          tcpTable[numTcpConnections].remIp = 
                                t[i].tcpConnRemAddress;

          tcpTable[numTcpConnections].remPort = 
                                (u_short)t[i].tcpConnRemPort;

          /* already in mib2 std values */
          tcpTable[numTcpConnections].state = 
                                t[i].tcpConnState;

          if (debugLevel) {
             printf("tcpConn: %lx:%d <-> %lx:%d\n",
      	            tcpTable[numTcpConnections].localIp,
	            tcpTable[numTcpConnections].localPort,
	            tcpTable[numTcpConnections].remIp,
	            tcpTable[numTcpConnections].remPort);
          }
          numTcpConnections++;
          enlarge_tcp_table();

       } /* if localPort != 0 */
       i++;
     }

     tcpTableTime = time(NULL);

     DBG1("convert_tcptable: %d total entries now\n",numTcpConnections);
}

/* ************************************************************* */
/* ************************************************************* */

u_char convertUDPState(u_char sockState)
{
     switch (sockState) {
       case MIB2_UDP_unbound:
	  return udpStateUnbound;
          break;
       case MIB2_UDP_idle:
	  return udpStateIdle;
          break;
       case MIB2_UDP_connected:
	  return udpStateConnected;
          break;
       case MIB2_UDP_unknown:
       default:
	  DBG1("udpState %d\n",sockState);
          return udpStateUnknown;
     }
}

static void convert_udp6table(mib2_udp6Entry_t *u6, int buflen)
{
     int numEntries,i;

     numEntries = buflen / sizeof(mib2_udpEntry_t);

     enlarge_udp_table();

     DBG1("convert_udp6table: %d entries\n",numEntries);

     i = 0;
     while (i < numEntries) {

       /* if localPort != 0 or ue_state is in bounds, then its a
          valid entry */

       if (((u_short)u6[i].udp6LocalPort != 0) && 
           (u6[i].udp6EntryInfo.ue_state <= MIB2_UDP_unknown) &&
           (u6[i].udp6EntryInfo.ue_state >= MIB2_UDP_unbound) ) {

	  /* ports in host byte order */

          udpTable[numUdps].ipVersion = 6;

          convertIPv6Addr(&u6[i].udp6LocalAddress,
                          udpTable[numUdps].localIPv6,48);
          udpTable[numUdps].localPort = (u_short)u6[i].udp6LocalPort;

          convertIPv6Addr(&u6[i].udp6EntryInfo.ue_RemoteAddress,
                          udpTable[numUdps].remIPv6,48);
          udpTable[numUdps].remPort = u6[i].udp6EntryInfo.ue_RemotePort;

          /* state of udp endpoint */
          udpTable[numUdps].state = 
               convertUDPState(u6[i].udp6EntryInfo.ue_state);

          if (debugLevel) {
             printf("udp6Entry: state:%d %lx:%d -> %lx:%d\n",
                 udpTable[numUdps].state,
                 udpTable[numUdps].localIp,
                 udpTable[numUdps].localPort,
                 udpTable[numUdps].remIp,
                 udpTable[numUdps].remPort);
          }

          numUdps++;
          enlarge_udp_table();

       } /* if local port != 0 */

       i++;
     }

     udpTableTime = time(NULL);

     DBG1("convert_udp6table: %d total entries now\n",numUdps);
}

static void convert_udptable(mib2_udpEntry_t *u, int buflen)
{
     int numEntries,i;

     numEntries = buflen / sizeof(mib2_udpEntry_t);

     enlarge_udp_table();

     DBG1("convert_udptable: %d entries\n",numEntries);

     i = 0;
     while (i < numEntries) {

       if ((u_short)u[i].udpLocalPort != 0) {
          udpTable[numUdps].ipVersion = 4;

          /* ports in host byte order according to mib2.h */

          udpTable[numUdps].localIp = u[i].udpLocalAddress;
          udpTable[numUdps].localPort = (u_short)u[i].udpLocalPort;

          udpTable[numUdps].remIp = u[i].udpEntryInfo.ue_RemoteAddress;
          udpTable[numUdps].remPort = u[i].udpEntryInfo.ue_RemotePort;

          /* state of udp endpoint */
          udpTable[numUdps].state = 
               convertUDPState(u[i].udpEntryInfo.ue_state);

          if (debugLevel) {
             printf("udpEntry: state:%d %lx:%d -> %lx:%d\n",
                 udpTable[numUdps].state,
                 udpTable[numUdps].localIp,
                 udpTable[numUdps].localPort,
                 udpTable[numUdps].remIp,
                 udpTable[numUdps].remPort);
          }

          numUdps++;
          enlarge_udp_table();

       } /* if local port != 0 */
       i++;
     }

     udpTableTime = time(NULL);

     DBG1("convert_udptable: %d total entries now\n",numUdps);
}

/* ************************************************************* */
/* ************************************************************* */

static void convert_routetable(mib2_ipRouteEntry_t *r, int buflen) 
{
     int i,ire_type;
     int numSolarisRoutes;
     char aStr1[256],aStr2[256],aStr3[256];

     enlarge_route_table();

     numRoutes = 0;  /* number of converted routes */
     i = 0;          /* index into array of routes to be converted */
     numSolarisRoutes = buflen / sizeof(mib2_ipRouteEntry_t);

     DBG2("convert_routetable: solaris routes, %d, %d bytes\n",
            numSolarisRoutes,buflen);

     /* for some reason, there is a lot of "junk" entries
        in the solaris routing table -- duplicates, etc. 
        netstat -rn only shows a small subet of whats often
        there; normally, netstat only shows interface, host, network 
        and default routes are shown; if you use netstat -rna,
        you can see all entries
 
        Some are broadcast and some are arp entries; we should get rid
        of arp and broadcast entries; through trial and error,
        the entries that show up in standard netstat-rn are
        IRE types 0x8, 0x2, and 0x80
     */
 
     while (i < numSolarisRoutes) {

       /* if its not a gateway entry, dont bother */
       ire_type = r[i].ipRouteInfo.re_ire_type;
       if ( ((ire_type & IRE_LOOPBACK) == 0) && 
            ((ire_type & IRE_DEFAULT) == 0) && 
            ((ire_type & IRE_IF_RESOLVER) == 0) ) {

	  i++;
          continue;
       }

       /* zero-terminate string if its not already done */
       r[i].ipRouteIfIndex.o_bytes[r[i].ipRouteIfIndex.o_length] = 0;
       sprintf(routeTable[numRoutes].ifname,"%s",
               r[i].ipRouteIfIndex.o_bytes);

       routeTable[numRoutes].ipv4dest = (u_int)r[i].ipRouteDest;
       routeTable[numRoutes].ipv4gway = (u_int)r[i].ipRouteNextHop;
       routeTable[numRoutes].ipv4mask = (u_int)r[i].ipRouteMask;
       routeTable[numRoutes].routeMetric = r[i].ipRouteMetric1;
       routeTable[numRoutes].routeType = r[i].ipRouteType;
       routeTable[numRoutes].routeProto = r[i].ipRouteProto;
       routeTable[numRoutes].routeAge = r[i].ipRouteAge;

       routeTable[numRoutes].flags = r[i].ipRouteInfo.re_flags;
       routeTable[numRoutes].refcnt = r[i].ipRouteInfo.re_ref;
       routeTable[numRoutes].use = r[i].ipRouteInfo.re_ref;
       /* see inet/ip.h for IRE types */
       routeTable[numRoutes].iretype = r[i].ipRouteInfo.re_ire_type;

       if (debugLevel) {

          xmpConvertIpAddr(routeTable[numRoutes].ipv4dest,aStr1,255);
          xmpConvertIpAddr(routeTable[numRoutes].ipv4gway,aStr2,255);
          xmpConvertIpAddr(routeTable[numRoutes].ipv4mask,aStr3,255);

          printf("Route: flags:%x ire:%lx dest: %s gway:%s mask:%s if:%s\n",
	      routeTable[numRoutes].flags,
	      routeTable[numRoutes].iretype,
	      aStr1,aStr2,aStr3,
	      routeTable[numRoutes].ifname);

       }

       numRoutes++;
       i++;

       enlarge_route_table();

     }
}

/* ************************************************************* */
/* not yet needed */

static void reloadIPv6Stats()
{
     u_char buf[32767];
     u_char ctrlbuf[1024];
     struct strbuf strbuf;
     struct strbuf datastr;
     struct T_optmgmt_req *treq;
     struct T_optmgmt_ack *tack;
     struct T_error_ack *teack;
     struct opthdr *opthdr;
     int flags,ret;
     static int s = 0;

     if (s == 0) {
        if ((s = open("/dev/ip6",O_RDWR,0600)) < 0) {
	   fprintf(stderr,"open of /dev/ip6 failed\n");
           return;
        }
     }

     strbuf.buf = (char *)ctrlbuf;
     strbuf.maxlen = sizeof(ctrlbuf);
     strbuf.len = sizeof(struct opthdr) + sizeof(struct T_optmgmt_req);

     treq = (struct T_optmgmt_req *)ctrlbuf;
     treq->PRIM_type = T_OPTMGMT_REQ;
     treq->OPT_offset = sizeof(struct T_optmgmt_req);
     treq->OPT_length = sizeof(struct opthdr);
     treq->MGMT_flags = T_CURRENT;

     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_req)];
     opthdr->level = MIB2_IP;  /* MIB group we want */
     opthdr->name = 0;         /* subgroup */
     opthdr->len = 0;

     flags = 0;
     if ((ret = putmsg(s,&strbuf,NULL,flags)) < 0) {
        fprintf(stderr,"reloadStats: putmsg failed\n");
        return;
     }

     /* on reply: strbuf ptr-> ctrlbuf[ containing t_optmgmt_ack opthdr ] */
     tack = (struct T_optmgmt_ack *)ctrlbuf;
     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_ack)];
     teack = (struct T_error_ack *)ctrlbuf;

     /* read replies until no more data */
     while (1) {

       /* read control data out of stream */
       flags = 0;  
       if ((ret = getmsg(s,&strbuf,NULL,&flags)) < 0) {
	  fprintf(stderr,"reloadStats: getmsg failed\n");
          break;
       }
       
       /* ack but no data, we're all done */ 
       if ((ret == 0) && (opthdr->len == 0) && 
           (tack->PRIM_type == T_OPTMGMT_ACK)) {
 	  DBG("reloadStats: ack, no data, were done\n");
 	  break; 
       }

       /* some sort of error */
       if ((teack->PRIM_type == T_ERROR_ACK) && 
           (strbuf.len >= sizeof(struct T_error_ack))) {
	 DBG2("reloadStats: getmsg returned errors %d %d\n",
                teack->TLI_error, teack->UNIX_error);
         break;
       }

       /* no more data to read, we're done */
       if ((ret | MOREDATA) == 0) {
	 DBG("reloadStats: no more data\n");
	 break;
       }

       switch (opthdr->level) {
         case MIB2_IP6:
	    printf("reloadIPv6Stats: mib2_ip6 level\n");
	    switch (opthdr->name) {
	      default:
                 printf("mib2_ip6 %d\n",opthdr->name);
                 flags = 0;
                 /* assert(opthdr->len < sizeof(buf)) !!! */
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len; 
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                    perror("getmsg of IP6 failed");
                 } 
	 	 break;        
            } /* mib2_ip6 */
            break;

         default:
	    printf("reloadIPv6Stats: level %d\n",opthdr->level);
            /* read the data (only exact amount!) and get it out of way */
	    /* assert(opthdr->len < sizeof(buf)) !!! */
            flags = 0;
            assert(opthdr->len < sizeof(buf));
            datastr.maxlen = opthdr->len;
            datastr.maxlen = sizeof(buf);
            datastr.buf = (char *)buf;
            datastr.len = 0;
            ret = getmsg(s,NULL,&datastr,&flags);
 	    break;

       }

     } /* while */

     return;
}

/* ************************************************************* */
/* reload stats for ip, icmp, tcp, udp via Solaris mib/sockoption */

static void reloadStats()
{
     /* u_char buf[32767]; */
     /* u_char ctrlbuf[4096]; */
     static int32_t buf[8192];
     static int32_t ctrlbuf[1024]; 
     struct strbuf strbuf;
     struct strbuf datastr;
     struct T_optmgmt_req *treq;
     struct T_optmgmt_ack *tack;
     struct T_error_ack *teack;
     struct opthdr *opthdr;
     int flags,ret;
     mib2_icmp_t *icmpptr;
     mib2_ip_t *ipptr;
     mib2_tcp_t *tcpptr;
     mib2_udp_t *udpptr;
     mib2_ipRouteEntry_t *r;
     mib2_ipAddrEntry_t *a;
     mib2_ipv6AddrEntry_t *a6;
     mib2_ipNetToMediaEntry_t *n;
     mib2_tcpConnEntry_t *t;
     mib2_tcp6ConnEntry_t *t6;
     mib2_udpEntry_t *u;
     mib2_udp6Entry_t *u6;

     /* we dont lock these stats structs since we are updating
        field at a time and there is no real harm if we are writing
        to a field whilst someone is reading from it */

     if (debugLevel)
        printf("reloadStats: starting\n");

     /* we dont lock these stats structs since we are updating
        field at a time and there is no real harm if we are writing
        to a field whilst someone is reading from it */

     /* strbuf ptr-> ctrlbuf[ containing T_optmgmt_req opthdr ] */

     strbuf.buf = (char *)ctrlbuf;
     strbuf.maxlen = sizeof(ctrlbuf);
     strbuf.len = sizeof(struct opthdr) + sizeof(struct T_optmgmt_req);

     treq = (struct T_optmgmt_req *)ctrlbuf;
     treq->PRIM_type = T_OPTMGMT_REQ;
     treq->OPT_offset = sizeof(struct T_optmgmt_req);
     treq->OPT_length = sizeof(struct opthdr);
     treq->MGMT_flags = T_CURRENT;

#if 0
     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_req)];
#endif
     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_req)/sizeof(int32_t)];
     opthdr->level = MIB2_IP;  /* MIB group we want */
     opthdr->name = 0;         /* subgroup */
     opthdr->len = 0;

     flags = 0;
     if ((ret = putmsg(s,&strbuf,NULL,flags)) < 0) {
        fprintf(stderr,"reloadStats: putmsg failed\n");
        return;
     }

     if (debugLevel)
        printf("reloadStats: putmsg returned %d\n",ret);

     /* on reply: strbuf ptr-> ctrlbuf[ containing t_optmgmt_ack opthdr ] */
     tack = (struct T_optmgmt_ack *)ctrlbuf;

     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_ack)];
     opthdr = (struct opthdr *)&ctrlbuf[sizeof(struct T_optmgmt_ack)/sizeof(int32_t)];

     teack = (struct T_error_ack *)ctrlbuf;

     /* lock the table structures since we are now getting
        udp6, tcp6 and perhaps other ipv6 tables */
     pthread_mutex_lock(&mib2_lock);

     /* reset number of UDP entries and tcp entries now; we do this
        because we want to combine udp and udp6 entries in a single
        table and we want to combine tcp and tcp6 entries in
        another single table but the endpoints can come out
        in different orders; that is, udp6 can come out before udp
        or after, and same for tcp.  Since we do not know a priori
        which order they will come out, we have to be careful
        not to overwrite entries
     */
     numUdps = 0;
     numTcpConnections = 0;
     numAddrs = 0;

     /* read replies until no more data */
     while (1) {

       /* control and data come out in parallel; need to be 
          careful that we read out only matching data for each
          control message we get as the data is often clumped
          together */

       /* read control data out of stream */
       flags = 0;  
       if ((ret = getmsg(s,&strbuf,NULL,&flags)) < 0) {
	  fprintf(stderr,"reloadStats: getmsg failed\n");
          break;
       }
       
       /* ack but no data, we're all done */ 
       if ((ret == 0) && (opthdr->len == 0) && 
           (tack->PRIM_type == T_OPTMGMT_ACK)) {
	  DBG("reloadStats: ack, no data, were done\n");
 	  break; 
       }

       /* some sort of error */
       if ((teack->PRIM_type == T_ERROR_ACK) && 
           (strbuf.len >= sizeof(struct T_error_ack))) {
	 fprintf(stderr,"reloadStats: getmsg returned errors %d %d\n",
                teack->TLI_error, teack->UNIX_error);
         break;
       }

       /* no more data to read, we're done */
       if ((ret | MOREDATA) == 0) {
	 DBG("reloadStats: no more data\n");
	 break;
       }

#if 0
       printf("returned level %d, name %d, len:%d\n",
               opthdr->level,opthdr->name,opthdr->len);
#endif

       switch (opthdr->level) {
         case MIB2_IP: 
            switch (opthdr->name) {
	      case 0:
                 flags = 0;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 assert(opthdr->len < sizeof(buf));
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
	 	    fprintf(stderr,"getmsg of ip stats failed\n");
                 }
                 else {
		    ipptr = (mib2_ip_t *)buf;
                    convert_ipstats(ipptr,&ipstats);
                 }
                 break;

	      case MIB2_IP_ROUTE:
                 DBG1("routetable: %d bytes\n",opthdr->len);
                 flags = 0;
                 /* XXX we're going to need to convert to a system
                    where we loop to take out data so as to avoid tripping
                    these assertions based on buffer lengths */
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of IP route failed");
                 }
                 else {
		    /* we've got route table in buf; dig entries out */
		    DBG1("routetable: %d bytes\n",datastr.len);
                    r = (mib2_ipRouteEntry_t *)buf;
                    convert_routetable(r,datastr.len);
                 }
		 break;


	      case MIB2_IP_ADDR:
                 DBG1("addrtable: %d bytes\n",opthdr->len);
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of ipAddrTable failed");
                 }
                 else {
		    /* we've got ipAddr table in buf; dig entries out */
		    DBG1("addrTable: %d bytes\n",datastr.len);
                    a = (mib2_ipAddrEntry_t *)buf;
                    convert_addrtable(a,datastr.len);
                 }
 	 	 break;

 	      case MIB2_IP_MEDIA: 
   	         DBG1("mediaTable len:%d\n",opthdr->len);
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of ipNetMediaTable failed");
                 }
                 else {
		    /* we've got ipNetMedia table in buf; dig entries out */
		    DBG2("mediaTable: %d bytes, flags %d\n",
                           datastr.len,flags);
                    DBG1("mediaTable: opthdr-len is %d\n",opthdr->len);
                    n = (mib2_ipNetToMediaEntry_t *)buf;
                    convert_mediatable(n,datastr.len);
                 }
 	 	 break;
           
  	      default: /* all others we want out of way */
	 	 DBG1("unsupported ip level opthdr %d\n",opthdr->name);
                 flags = 0;
                 /* assert(opthdr->len < sizeof(buf)) !!! */
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len; 
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                    perror("getmsg of IP failed");
                 } 
	 	 break;        
   
            } /* switch IP opthdr->name under MIB2_IP */
            break;

         case MIB2_TCP6:
	    DBG1("TCP6 . . . %d\n",opthdr->name);
 	    switch (opthdr->name) {
	      case MIB2_TCP6_CONN:
	 	 /* get the tcp6 connection entries */
		 DBG1("TCP6 connections, len %d\n",opthdr->len);
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of tcp6Table failed");
                 }
                 else {
		    /* we've got tcp table in buf; dig entries out */
		    DBG1("tcp6Table: %d bytes\n",datastr.len);
                    t6 = (mib2_tcp6ConnEntry_t *)buf;
                    convert_tcp6table(t6,datastr.len);
                 }
 	 	 break;

	      default: /* all others we want out of way */
           	 DBG1("TCP6, name %d\n",opthdr->name);
                 flags = 0;
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
		    perror("getmsg of tcp6  failed");
                 }
            } /* switch TCP6 opthdr->name */
            break;

         case MIB2_IP6:
	    DBG1("reloadStats: IP6 . . .%d\n",opthdr->name);
	    switch (opthdr->name) {
              case MIB2_IP6_ADDR:
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                    perror("getmsg of ipv6AddrTable failed");
                 }
                 else {
		    /* we've got ipv6Addr table in buf; dig entries out */
                    a6 = (mib2_ipv6AddrEntry_t *)buf;
                    convert_addr6table(a6,datastr.len);
                 }
 	         break;

	      case MIB2_IP6_ROUTE:
	      case MIB2_IP6_MEDIA:
	      case EXPER_IP6_GROUP_MEMBERSHIP:

  	      default: /* all others we want out of way */
                 flags = 0;
                 /* assert(opthdr->len < sizeof(buf)) !!! */
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len; 
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                    perror("getmsg of IP6 failed");
                 } 
	 	 break;        
            } /* switch IP6 opthdr->name */
            break;

         case MIB2_UDP6:
	    DBG1("UDP6 . . .%d\n",opthdr->name);
 	    switch (opthdr->name) {

	      case MIB2_UDP6_ENTRY:
	 	 /* get the udp6 endpoints */
	 	 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 assert(opthdr->len < sizeof(buf));
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
	 	    perror("getmsg of udp6Table failed");
                 }
                 else {
	 	    /* we've got udp6 table in buf6; dig out entries */
	 	    DBG1("udp6Table: %d bytes\n",datastr.len);
                    u6 = (mib2_udp6Entry_t *)buf;
                    convert_udp6table(u6,datastr.len);
                 }
	 	 break;

	      default: /* all others we want out of way */
           	 DBG1("UDP6, name %d\n",opthdr->name);
                 flags = 0;
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
		    perror("getmsg of udp6 failed");
                 }
            } /* switch TCP6 opthdr->name */
            break;

         case MIB2_ICMP:
	    /* right now, we only handle name 0: */
	    /* read out mib2_icmp_t struct */
            flags = 0;
            datastr.maxlen = sizeof(mib2_icmp_t);
            datastr.buf = (char *)buf;
            datastr.len = 0;
            if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
	       perror("getmsg of icmp stats failed");
            }
            else {
	       icmpptr = (mib2_icmp_t *)buf;
               convert_icmpstats(icmpptr,&icmpstats);
            }
 	    break;

         case MIB2_TCP:
	    switch (opthdr->name) {
	      case 0:
                 flags = 0;
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 assert(opthdr->len < sizeof(buf));
                 /* assert(opthdr->len < sizeof(buf)) !!! */
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
	 	    perror("getmsg of tcp stats failed");
                 }
                 else {
		    tcpptr = (mib2_tcp_t *)buf;
                    convert_tcpstats(tcpptr,&tcpstats);
                 }
                 break;

	      case MIB2_TCP_CONN:
		 DBG1("TCP connections, len %d\n",opthdr->len);
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of tcpTable failed");
                 }
                 else {
		    /* we've got tcp table in buf; dig entries out */
		    DBG1("tcpTable: %d bytes\n",datastr.len);
                    t = (mib2_tcpConnEntry_t *)buf;
                    convert_tcptable(t,datastr.len);
                 }
 	 	 break;

	      default: /* all others we want out of way */
           	 DBG1("TCP, name %d\n",opthdr->name);
                 flags = 0;
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
		    perror("getmsg of tcp failed");
                 }

	    } /* tcp opthdr->name */

 	    break;

         case MIB2_UDP:
	    switch (opthdr->name) {
	      case 0:
                 flags = 0;
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 assert(opthdr->len < sizeof(buf));
                 /* assert(opthdr->len < sizeof(buf)) !!! */
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
	 	    fprintf(stderr,"getmsg of udp stats failed");
                 }
                 else {
		    udpptr = (mib2_udp_t *)buf;
                    convert_udpstats(udpptr,&udpstats);
                 }
                 break;

	      case MIB2_UDP_ENTRY:
		 DBG1("udpTable: %d bytes\n",opthdr->len);
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 assert(opthdr->len < sizeof(buf));
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
                     perror("getmsg of udpTable failed");
                 }
                 else {
		    /* we've got udp table in buf; dig entries out */
		    DBG1("udpTable: %d bytes\n",datastr.len);
                    u = (mib2_udpEntry_t *)buf;
                    convert_udptable(u,datastr.len);
                 }
                 break;

	      default: /* others we want out of way */
                 DBG1("UDP, name %d\n",opthdr->name);
                 flags = 0;
                 assert(opthdr->len < sizeof(buf));
                 datastr.maxlen = opthdr->len;
                 datastr.maxlen = sizeof(buf);
                 datastr.buf = (char *)buf;
                 datastr.len = 0;
                 if ((ret = getmsg(s,NULL,&datastr,&flags)) < 0) {
		    perror("getmsg of udp failed");
                 }
              
            } /* switch udp opthdr->name */
            break;

         /* ones we're not interested in (yet) */
         case MIB2_INTERFACES: 
	    printf("reloadStats: mib2 interfaces!\n");

         case MIB2_ICMP6: /* need to add this ? XXX */

         case MIB2_SYSTEM:
         case MIB2_AT:
         case MIB2_EGP:
	 case MIB2_CMOT:
	 case MIB2_TRANSMISSION:
	 case MIB2_SNMP:
         default:
	    DBG1("reloadStats: other level %d\n",opthdr->level);
            /* read the data (only exact amount!) and get it out of way */
	    /* assert(opthdr->len < sizeof(buf)) !!! */
            flags = 0;
            assert(opthdr->len < sizeof(buf));
            datastr.maxlen = opthdr->len;
            datastr.maxlen = sizeof(buf);
            datastr.buf = (char *)buf;
            datastr.len = 0;
            ret = getmsg(s,NULL,&datastr,&flags);
 	    break;
        
       } /* switch level */

     } /* while more data to read out of socket s */
    
     statsTime = time(NULL);

     /* release mib2 lock */
     pthread_mutex_unlock(&mib2_lock);

     if (debugLevel)
        printf("reloadStats: done\n");

     return;
}

/* ************************************************************* */

static void *solarisMib2_dowork(void *arg)
{
     time_t current;
     int i;

     DBG("solarisMib2_dowork: starting\n");

     while (mib2ThreadKill != 1) {

       current = time(NULL);

       if ((current - statsTime) > MIB2_CACHE_LIFE) {
  	  /* this function reloads everything but interface stats */
	  reloadStats(); 
       }

       /* if table */
       if ((current - ifTableTime) > MIB2_CACHE_LIFE) {
	  reloadIfTable();
       }

       for (i=0; i<2 && mib2ThreadKill == 0; i++) {
           sleep(15);
       }

       DBG("solarisMib2_dowork: working . . .\n");

     } /* while 1 */

     printf("solarisMib2: worker thread returning\n");

     return NULL;  /* effectively killing me */
}

/* ************************************************************* */

int start()
{
     int mib2TrapHandler();

     printf("solarisMib2: start invoked\n");

     /* open /dev/ip for using T_OPTMGMT_REQ ioctl */
     /* push on TCP, UDP, ARP */
     if ((s = open("/dev/ip",O_RDWR,0600)) < 0) {
        fprintf(stderr,"open of /dev/ip failed\n");
     }

     /* very important to push arp module before others XXX */
     if (ioctl(s,I_PUSH,"arp") < 0) {
        fprintf(stderr,"mib2: ioctl push arp failed\n");
     }

     if (ioctl(s,I_PUSH,"tcp") < 0)
        fprintf(stderr,"mib2: ioctl push tcp failed\n");

     if (ioctl(s,I_PUSH,"udp") < 0)
        fprintf(stderr,"mib2: ioctl push udp failed\n");

     DBG1("solarisMib2: sizeof(longlong_t) is %d\n",sizeof(longlong_t));

     /* start a thread */
     if (pthread_create(&wt,NULL,solarisMib2_dowork,NULL) < 0) {
        fprintf(stderr,"Failed to create worker thread for solarisMib2\n");
        return -1;
     }
     /* detach thread */
     pthread_detach(wt);

     xmpRegisterTrapHandler("mib2",mib2TrapHandler);

     return 1;
}

/* ************************************************************* */

#endif /* SOLARIS */

