/* 
   COPYRIGHT 2006 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - coreEvents.c */
/* rdk@krupczak.org */

/* core MIB events table and related code */

#ifdef WIN32
#include <windows.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>

#ifndef WIN32
#include <strings.h>
#include <arpa/inet.h>
#include <sys/utsname.h>
#endif

#include "xmp.h"
#include "xmp_vars.h"
#include "linkedlist.h"
#include "xmp_session.h"
#include "xmp_events.h"
#include "globals.h"
#include "pluginexport.h"
#include "socklib.h"

static linkedList_t *eventList = NULL;
static linkedList_t *eventRegistryList = NULL;
extern xmpStats_t xmpStats;
extern struct utsname systemuts;

char *base64_encode(u_char *, int, int *);

/* ************************************************************* */

static int isThisMe(char *nameaddr)
{
     if (nameaddr == NULL) {
        return 0;
     }

     /* get rid of some easy ones first! */
     if (nameaddr[0] == 0) {
        return 1;
     }
     if (strstr(nameaddr,"localhost") != NULL) {
        return 1;
     }
     if (strcmp(nameaddr,"127.0.0.1") == 0) {
        return 1;
     }
     /* we dont want 0.0.0.0 */
     if (strcmp(nameaddr,"0.0.0.0") == 0) {
        return 1; 
     }
     /* IPv6 local host addresses */
     if (strcmp(nameaddr,"::1") == 0) {
        return 1; 
     }
     if (strcmp(nameaddr,"::1") == 0) {
        return 1; 
     }
     if (strcmp(nameaddr,"0000:0000:0000:0000:0000:0000:0000:0000") == 0) {
        return 1; 
     }
     if (strcmp(nameaddr,"0000:0000:0000:0000:0000:0000:0000:0001") == 0) {
        return 1; 
     }

     if (strcmp(nameaddr,systemuts.nodename) == 0) {
        return 1;
     }
     if (strstr(nameaddr,systemuts.nodename) != NULL) {
        return 1;
     }
     if (strstr(systemuts.nodename,nameaddr) != NULL) {
        return 1;
     }

     return 0;
}

/* ************************************************************* */

int eventCompare(xmpEventEntry_t *a, xmpEventEntry_t *b)
{
     if ((a == NULL) || (b == NULL))
        return -1;

     if (a == b)
        return 0;

     return 1;
}

int eventPrint(xmpEventEntry_t *e)
{
     if (e == NULL)
        return -1;

     fprintf(stderr,"Event: src:'%s' type:'%s' sysobjid:%s\n",
             e->eventSource,e->eventType,e->sysObjectID);

     return 1;
}

int eventFree(xmpEventEntry_t *e)
{
     xmpPdu_t *eventPdu;

     if (e == NULL)
        return -1;

     /* free the PDU associated with the event */
     eventPdu = e->eventPdu;
     if (eventPdu != NULL) {
        xmpPduCleanup(eventPdu);
        bzero((void *)eventPdu,sizeof(xmpPdu_t));
        free(eventPdu);
     }

     if (e->eventPduXml != NULL) {
        free(e->eventPduXml);
     }

     /* free the event entry itself */
     bzero((void *)e,sizeof(xmpEventEntry_t));
     free(e);     

     return 1;
}

/* ************************************************************* */
/* receive a trap via network and put it in event table;
   processEvents will dispatch the trap to interested parties */

xmpErrorStatus_t receiveTrap(char *ipaddr, xmpPdu_t *in, 
                             u_char *buffer, int len)
{
     xmpEventEntry_t *e;
     xmpPdu_t *copy;
     xmpVariable_t *v;

     /* alloc an event structure and copy the PDU since the original PDU
        is destroyed by the session when it closes */

     e = (xmpEventEntry_t *)malloc(sizeof(xmpEventEntry_t));
     if (e == NULL) {
        fprintf(stderr,"receiveTrap: failed to malloc space for event\n");
        return resourceUnavailable;
     }
     bzero((void *)e,sizeof(xmpEventEntry_t));
 
     if ((copy = (xmpPdu_t *)malloc(sizeof(xmpPdu_t))) == NULL) {
        free(e);
        fprintf(stderr,"receiveTrap: failed to malloc space for event\n");
        return resourceUnavailable;
     }
     bzero((void *)copy,sizeof(xmpPdu_t));

     /* copy the PDU *and* all the variables */
     copyXmpPdu(in,copy);

     sprintf(e->eventSource,"%s",ipaddr);
     e->timeStamp = time(NULL);
     e->ageOut = eventAgeOutInterval;
     e->dispatched = 0;
     e->eventPdu = copy;

     /* get the event type and sysObjectID out of trap */
     v = copy->varList;
     while (v != NULL) {

       if ((strcmp(v->varName.mibName,"core") == 0) &&
	   (strcmp(v->varName.objName,"trapType") == 0)) {

	  strncpy(e->eventType,v->valueStr,255);
          e->eventType[255] = 0;

       }

       if ((strcmp(v->varName.mibName,"core") == 0) &&
	   (strcmp(v->varName.objName,"sysObjectID") == 0)) {

	  strncpy(e->sysObjectID,v->valueStr,255);
          e->sysObjectID[255] = 0;
       }

       v = v->next;
     }

     /* copy the raw xml-encoded trap PDU for later */
     e->xmlSize = 0;
     e->eventPduXml = (char *)malloc(len+2);
     if (e->eventPduXml != NULL) {
        e->xmlSize = len; /* length w/o zero-termination */
        strncpy(e->eventPduXml,buffer,len);
        e->eventPduXml[len] = 0;
     }

     insertHead(eventList,(void *)e);

     DBG1("receiveTrap: received trap from %s\n",ipaddr);

     /* for each member of the trap registry, iterate through list
        and see if they want this trap XXX */
     
     return noError;
}

/* ************************************************************* */

xmpErrorStatus_t receiveInformation(char *ipaddr, xmpPdu_t *in, 
                                    u_char *buffer, int len)
{

     /* dispatch information to whoever would like it */

     return noError;
}

/* ************************************************************* */
/* send a pre-formulated trap PDU to the system with IP address
   or hostname; caller creates Trap PDU and vars but we free it after 
   event ageout */

xmpErrorStatus_t sendTrap(char *ipaddr, xmpPdu_t *t, char *eventType, 
                          char *sysObjID, char *authenUser)
{
     linkedElement_t *l;
     xmpEventEntry_t *e;
     xmpSession_t *session;
     char hostname[256]; /* who we are sending to */
     u_int ipv4addr;
     xmpPdu_t reply;
     int ret;
     char *dest;
     int needed = 0, bufSize;

     DBG("sendTrap: starting\n");

     if ((t == NULL) || (eventType == NULL) || (sysObjID == NULL)) {
        fprintf(stderr,"sendTrap: invalid args\n");
        return badValue;
     }

     dest = ipaddr;
     if ((ipaddr == NULL) || (ipaddr[0] == 0)) {
        dest = defaultManager;
     }

     DBG("sendTrap: going to see if trap already on list\n");

     /* if the trap pdu is already on the list, dont re-add it */
     /* lock the linked list */
     pthread_mutex_lock(&eventList->lock);

     l = eventList->head;

     while (l != NULL) {

       e = (xmpEventEntry_t *)l->data;
       if (e == NULL) {
	  l = l->next;
          continue;
       }

       if (e->eventPdu == t) {
	  break;
       }

       l = l->next;
     }

     /* dont keep lock on event list so as to avoid race conditions */ 
     pthread_mutex_unlock(&eventList->lock);

     DBG1("sendTrap: looked for event on list, %lx\n",l);

     /* if l is null, we did not find the trap PDU already on list */
     if (l == NULL) {

        e = (xmpEventEntry_t *)malloc(sizeof(xmpEventEntry_t));
        if (e == NULL) {
           fprintf(stderr,"sendTrap: failed to malloc space for event\n");
           return resourceUnavailable;
        }
        bzero((void *)e,sizeof(xmpEventEntry_t));

        e->timeStamp = time(NULL);
        e->ageOut = eventAgeOutInterval;
        e->eventSource[0] = 0; /* locally originated */

        /* should we dispatch to ourselves ? XXX 
           by setting dispatched=0, we will cause the event to 
           be sent to ourselves. */

        e->dispatched = 0; 
        e->eventPdu = t;

        sprintf(e->eventType,"%s",eventType);
        sprintf(e->sysObjectID,"%s",sysObjID);

        /* xml-encode the event for future retrieval */
        e->eventPduXml = NULL;
        e->xmlSize = 0;
        bufSize = 2048;
        e->eventPduXml = (char *)malloc(bufSize); /* guessitmate */

        DBG2("sendTrap: pdu %lx eventPduXml %lx\n",
              e->eventPdu,e->eventPduXml);

        if (e->eventPduXml != NULL) {
           needed = 0;
	   ret = buildXmpMessage(e->eventPdu,e->eventPduXml,bufSize,&needed);
           if ((ret < 0) && (needed != 0)) {
	      e->eventPduXml = (char *)realloc(e->eventPduXml,needed+2);
              if (e->eventPduXml != NULL) {
		 bufSize = needed+1;
                 needed = 0;
   	         ret = buildXmpMessage(e->eventPdu,e->eventPduXml,bufSize,
                                       &needed);
              }
           }
           if (ret < 0) {
	      if (e->eventPduXml)
	         free(e->eventPduXml);
              e->eventPduXml = NULL;
              e->xmlSize = 0;
           }
           else {
	      e->xmlSize = ret;
           }

        } /* if eventPduXml != NULL */

        insertHead(eventList,(void *)e);
     }
     else {
        DBG("Trap already on list\n");
     }

     /* we dont cache which plugin or component sent the trap */

     DBG1("sendTrap: going to send to '%s'\n",defaultManager);

     /* make sure we're not sending to ourselves; */
     hostname[0] = 0;
     if (ipAddrToHostname(dest,hostname,255) < 0) {
        DBG("sendTrap: couldnt convert address to hostname\n");
        return noError;
     }

     if (isThisMe(hostname) != 0) {

        /* if its me, just make sure its added to our internal table 
           but dont send it to myself */

        DBG("sendTrap: should not send to myself\n");

        return noError;
     }

     DBG1("sendTrap: going to establish session with '%s'\n",hostname);

     if (hostname[0] == 0) {

        DBG("Warning: xmpSendTrap not sending to '' *********\n");

        return 0;
     }

     /* establish session with target system and then send trap */
     ipv4addr = xmpHostnameToIpaddr(hostname);

     if (ipv4addr == 0) {

       DBG1("Warning: xmpSendTrap should not send to 0, hostname '%s' *********\n",hostname);

        return 0;
     }

     if ((session = newXmpSessionConnectTimeout(ipv4addr,
               (u_short)htons((u_short)XMP_PORT),3,authenUser)) == NULL) {
        return 0;
     }

     bzero((void *)&reply,sizeof(reply));

     if ((ret = xmpSendQueryReceiveReply(session,t,&reply)) < 0) {

        DBG1("sendTrap: xmpSendQueryReceiveReply returned %d\n",ret);

        /* we should mark that the send failed and re-send? XXX */
        /* if so, we'd have to mark for every receiver */
     }

     destroyXmpSession(session);
    
     xmpPduCleanup(&reply);

     DBG1("sendTrap: finished sending trap to %s\n",hostname);

     return reply.errorStatus;
}

/* ************************************************************* */
/* initialize our event table and related functionality */

int initCoreEvents()
{

     eventList = newLinkedList(sizeof(xmpEventEntry_t),
                               eventCompare,
                               eventPrint,
                               eventFree,
                               "coreEvents",1);

     /* eventAgeOutInterval is either 0 or its set in the config file */

     eventRegistryList = newLinkedList(sizeof(xmpEventRegistry_t),
                                       NULL,NULL,NULL,"eventRegistry",1);

     return 0;
}

/* ************************************************************* */
/* periodically, go through event list and dispatch any events
   that may have arrived and have not yet been forwarded to
   interested parties; also, events that have expired, go ahead
   and eliminate */

int processEvents()
{
     linkedElement_t *l, *ptr;
     xmpEventEntry_t *e;
     time_t now;
     int numRegisters,i;
     xmpEventRegistry_t *r;

     DBG("processEvents: starting\n");

     /* lock the linked list */
     pthread_mutex_lock(&eventList->lock);

     l = eventList->head;
     now = time(NULL);

     while (l != NULL) {

       e = (xmpEventEntry_t *)l->data;
       if (e == NULL) {
	  l = l->next;
          continue;
       }

       /* if the ageout interval is non-zero, delete events after
          that many seconds have gone by */

       if ((eventAgeOutInterval != 0) && 
           ((now - e->timeStamp) > eventAgeOutInterval)) {

	  DBG2("Going to age out trap '%s' from '%s'\n",
               e->eventType,e->eventSource);

	  ptr = l;
          l = l->next;
          removeElement(eventList,ptr);
          continue;
       }

       /* should we have freed the linked element l ? No, 
          removeElement handles that plus free'ing up the
          PDU and PDU varbinds */

       if (e->dispatched == 0) {
	  /* dispatch the event to those who signed up for it */
	  DBG2("Examining event '%s' from '%s'\n",
                e->eventType,e->eventSource);

          /* get a copy of event registry list */
	 r = (xmpEventRegistry_t *)toArrayCopy(eventRegistryList,
                                               &numRegisters);
         if (r != NULL) {

	    for (i=0; i<numRegisters; i++) {

	        if (strstr(e->eventType,r[i].trapFilter) != NULL) {
	           DBG2("Dispatching event '%s' from '%s'\n",
                        e->eventType,e->eventSource);

		   if (r[i].trapHandler != NULL)
		      (r[i].trapHandler)(e);
		}

 	    } /* for each registry entry */

  	    e->dispatched = 1; 

            /* free up array copy of list */
            free(r);

         } /* if r != null, registry copied */

       } /* if e->dispatched == 0 */

       /* should we have freed the linked element l ? XXX 
          I dont think so because we are traversing the list but
          not removing from it */

       l = l->next;

     } /* for each event */

     pthread_mutex_unlock(&eventList->lock);

     DBG("processEvents: finished\n");

     return 1;
}

/* ************************************************************* */

xmpErrorStatus_t var_eventTable(char *key, xmpVariable_t *columns,
                                xmpPdu_t *out)
{
     char rowKey[1024];
     xmpVariable_t *varptr, *cptr;
     u_int maxrows;
     int numRows = 0;
     int numVariables = 0;
     char timestr[32];
     linkedElement_t *ptr;
     xmpEventEntry_t *e;
     static char *theMessageStr = NULL;
     static int msgStrSize = 0;

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     if (theMessageStr == NULL) {
        theMessageStr = (char *)malloc(32000);
        if (theMessageStr != NULL) {
	   msgStrSize = 32000;
        }
     }

     DBG("var_eventTable: starting\n");

     /* keys can be XXX, *, \0 */
     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     DBG("var_eventTable: going to lock\n");

     pthread_mutex_lock(&eventList->lock);

     DBG("var_eventTable: got lock\n");

     /* cycle through event table */
     ptr = eventList->head;
     while (ptr != NULL) {

       e = (xmpEventEntry_t *)ptr->data;
       if (e == NULL) {
	  ptr = ptr->next;
          continue;
       }

       /* row key is  . . . .*/
       xmpConvertTimeStr(e->timeStamp,timestr,31);       
#ifndef WIN32
       snprintf(rowKey,2047,"%s,%s,%s,%s",
#else
       _snprintf(rowKey,2047,"%s,%s,%s,%s",
#endif
               e->eventSource,
               timestr,
               e->eventType,
               e->sysObjectID);

       if ((strcmp(rowKey,key) != 0) && (key[0] != 0) && 
           (key[0] != '*')) {
	  ptr = ptr->next;
 	  continue; /* with events */
       }

       cptr = columns;
       while ((cptr != NULL) && (numRows < maxrows)) {

         /* malloc a return value */
         if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
             == NULL) {
	     perror("var_eventTable: malloc failed");
             out->errorStatus = genError;
             /* release lock */
             pthread_mutex_unlock(&eventList->lock);
             return genError;
         }
         bcopy(cptr,varptr,sizeof(xmpVariable_t));
         strcpy(varptr->varName.key,rowKey);

         switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

	   case 1: /* source */
	      varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            e->eventSource);
	      break;
	   case 2: /* timestamp */
	      varptr->varSyntax = displayString;
              xmpConvertTimeStr(e->timeStamp,timestr,sizeof(timestr));
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",timestr);
	      break;

	   case 3: /* type */
	      varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            e->eventType);
	      break;

	   case 4: /* sysObjID */
	      varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            e->sysObjectID);
	      break;

	   case 5: /* ageout */
	      varptr->varSyntax = gauge;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                            e->ageOut);
	      break;

  	   case 6: /* event PDU itself */
	      varptr->varSyntax = displayString;
              varptr->valueStrLen = 0;
              varptr->valueStr[0] = 0;
              if ((e->eventPduXml != NULL) && (e->xmlSize > 0)) {
	 	 varptr->valueBuf = base64_encode(e->eventPduXml,
                                                  e->xmlSize,
                                                  &varptr->valueStrLen);
              }
              break;

	   default:
	      fprintf(stderr,"var_eventTable: invalid column\n");
              /* release lock */
              pthread_mutex_unlock(&eventList->lock);
              out->errorStatus = noSuchName;
              out->errorIndex = numVariables;
              out->numberVariables = numVariables;
              return noSuchName;

         } /* switch column */

         /* glue into variable list */
         numVariables++;
         if (out->varListTail)
            out->varListTail->next = varptr;
         out->varListTail = varptr;
         if (out->varList == NULL)
            out->varList = varptr;           

         cptr = cptr->next;

       } /* while columns */

       numRows++;
       ptr = ptr->next;

     } /* while events */

     pthread_mutex_unlock(&eventList->lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (xmpd_verbose)
        printf("var_eventTable: %d variables, finished\n",numVariables);

     return noError;
}


/* ************************************************************* */
/* register to receive traps from a particular MIB (e.g. cartographer, mib2) */
/* traps can be as specific as a particular exact type or simply a mib name */

#ifndef WIN32
int registerTrapHandler(char *name, int (*fptr)())
#else
int registerTrapHandler(char *name, FARPROC fptr)
#endif
{
     xmpEventRegistry_t *er;

     if ((name == NULL) || (fptr == NULL))
        return -1;

     er = (xmpEventRegistry_t *)malloc(sizeof(xmpEventRegistry_t));
     if (er == NULL) {
        return -1;
     }

     bzero((void *)er,sizeof(xmpEventRegistry_t));
     strncpy(er->trapFilter,name,255);
     er->trapFilter[255] = 0;

     er->registerTime = time(NULL);

     er->trapHandler = fptr;

     insertHead(eventRegistryList,(void *)er);

     return 1;
}

/* ************************************************************* */

int sendColdstartTrap()
{
     xmpPdu_t *t;
     xmpVariable_t *v1,*v2;
     int ret;

     fprintf(stderr,"sendColdstartTrap: starting to '%s'\n",defaultManager);

     if ((t = (xmpPdu_t *)malloc(sizeof(xmpPdu_t))) == NULL) {
        return resourceUnavailable;
     }
     if ((v1 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return resourceUnavailable;
     }
     if ((v2 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        free(v1);
        return resourceUnavailable;
     }

     bzero((void *)t,sizeof(xmpPdu_t));
     t->msgType = Trap;
     t->messageID = (int64_t)time(NULL);
     t->errorStatus = noError;
     t->errorIndex = 0;

     t->numberVariables = 2;
     bzero((void *)v1,sizeof(xmpVariable_t));
     bzero((void *)v2,sizeof(xmpVariable_t));
     t->varList = v1;
     t->varListTail = v2;

     sprintf(v1->varName.mibName,"core");
     sprintf(v1->varName.objName,"trapType");
     v1->varSyntax = displayString;
     v1->valueStrLen = sprintf(v1->valueStr,"core.coldStart");
     v1->next = v2;

     sprintf(v2->varName.mibName,"core");
     sprintf(v2->varName.objName,"sysObjectID");
     v2->varSyntax = displayString;
     v2->valueStrLen = sprintf(v2->valueStr,KORGOBJID);
     v2->next = NULL;

     /* dont free traps nor variables since they are placed on
        event list and later cleaned up after ageout */

     ret = sendTrap(defaultManager, t, "core.coldStart",
		    KORGOBJID,NULL);

     return ret;
}

/* ************************************************************* */

